from typing import Iterable, Optional, Generator
from itertools import chain
from pathlib import Path

import toml

from .namespace import NameSpace
from .decl import StructDecl
from .parse import parse_file
from .shiftreduce import ParseError


__all__ = (
    'load',
)


def load(files: Iterable[Path],
         registry: Optional[Path] = None,
         module_name: Optional[str] = None,
         ) -> NameSpace:
    file_set = list(files)
    single = len(file_set) == 1

    if registry is not None:
        reg = toml.load(registry)
    else:
        reg = {}

    if module_name:
        module_name = module_name
    elif meta_name := reg.get('meta', {}).get('name'):
        module_name = meta_name
    elif single:
        in_path, = file_set
        module_name = in_path.stem

    def parse(p: Path) -> Generator[StructDecl, None, None]:
        try:
            yield from parse_file(p)
        except ParseError as e:
            print(e)
            raise SystemExit(1)

    def parse_with_prefix(p: Path) -> Generator[StructDecl, None, None]:
        name = p.stem

        if reg:
            module_id = reg['modules'][name]
            struct_map = reg['module'][name]
        else:
            module_id = 0
            struct_map = {}

        assert 0 <= module_id <= 0xffff

        for s in parse(p):
            decl = s.with_prefix(name)

            if reg:
                struct_name = s.struct_name
                struct_id = struct_map[struct_name]

                assert 0 <= struct_id <= 0xffff

                discriminant = (module_id << 16) | struct_id
                decl = decl.with_discriminant(discriminant)

            yield decl

    if single:
        in_file, = file_set
        decls = tuple(parse(in_file))
    else:
        decls = tuple(chain(*map(parse_with_prefix, file_set)))

    return NameSpace.from_decls(decls, name=module_name)



from typing import Dict
from enum import Enum

from .integraltype import IntegralType
from .buftypes import BufType, Utf8Type, ByteArrayType, StringArrayType
from .uuidtype import UuidType
from .intstack import IntStackType
from .typedef import TypeDef

__all__ = (
    'BaseType',
    'base_types',
)

Bytes = BufType()
Utf8 = Utf8Type()
ByteArray = ByteArrayType()
StringArray = StringArrayType()

UnsignedInt8 = IntegralType(8, False)
SignedInt8 = IntegralType(8, True)
UnsignedInt16 = IntegralType(16, False)
SignedInt16 = IntegralType(16, True)
UnsignedInt32 = IntegralType(32, False)
SignedInt32 = IntegralType(32, True)
UnsignedInt64 = IntegralType(64, False)
SignedInt64 = IntegralType(64, True)

Uuid = UuidType()

IntStack8 = IntStackType(8)
IntStack16 = IntStackType(16)
IntStack32 = IntStackType(32)


class BaseType(TypeDef, Enum):
    bytes = 'bytes', Bytes
    utf8 = 'utf8', Utf8
    bytearray = 'bytearray', ByteArray
    stringarray = 'stringarray', StringArray

    u8 = 'u8', UnsignedInt8
    i8 = 'i8', SignedInt8
    u16 = 'u16', UnsignedInt16
    i16 = 'i16', SignedInt16
    u32 = 'u32', UnsignedInt32
    i32 = 'i32', SignedInt32
    u64 = 'u64', UnsignedInt64
    i64 = 'i64', SignedInt64

    u128 = 'u128', Uuid
    uuid = 'uuid', Uuid

    intstack8 = 'intstack8', IntStack8
    intstack16 = 'intstack16', IntStack16
    intstack32 = 'intstack32', IntStack32


def base_types() -> Dict[str, TypeDef]:
    return {t.name: t.value for t in BaseType}



from typing import Iterable, TypeVar, List
from collections import Counter


__all__ = (
    'dupes',
)


_T = TypeVar('_T')


def dupes(s: Iterable[_T]) -> List[_T]:
    return [x for x, y in Counter(s).items() if y > 1]



from .type import XpdtType
from .integraltype import IntegralType

__all__ = (
    'BufType',
    'Utf8Type',
    'ByteArrayType',
    'StringArrayType',
)


class BufBase(XpdtType):
    __slots__ = (
        '_off_type',
    )

    def __init__(self) -> None:
        self._off_type = IntegralType(32, False)

    @property
    def size(self) -> int:
        return self._off_type.size

    @property
    def struct_fmt(self) -> str:
        return self._off_type.struct_fmt

    @property
    def ctype(self) -> str:
        return 'struct xbuf'

    @property
    def needs_vbuf(self) -> bool:
        return True

    @property
    def needs_decode(self) -> bool:
        return True


class BufType(BufBase):
    __slots__ = ()

    @property
    def pytype(self) -> str:
        return 'bytes'

    def read_func(self, s: str) -> str:
        return f'tobytes({s})'

    def write_func(self, s: str) -> str:
        return s


class Utf8Type(BufBase):
    __slots__ = ()

    @property
    def pytype(self) -> str:
        return 'str'

    def read_func(self, s: str) -> str:
        return f'tostr(tobytes({s}))'

    def write_func(self, s: str) -> str:
        return f'{s}.encode()'


class ByteArrayType(BufBase):
    __slots__ = ()

    @property
    def pytype(self) -> str:
        return '_Tup[int]'

    def read_func(self, s: str) -> str:
        return f'tuple({s})'

    def write_func(self, s: str) -> str:
        return f'bytes({s})'


class StringArrayType(BufBase):
    __slots__ = ()

    @property
    def pytype(self) -> str:
        return '_Tup[str]'

    def read_func(self, s: str) -> str:
        return f"tuple((x.decode() for x in tobytes({s}).split(b'\\x00')))"

    def write_func(self, s: str) -> str:
        return f"b'\\x00'.join((x.encode() for x in {s}))"



from typing import (
    Optional, Tuple, Dict, TextIO, Iterable, Iterator, Any
)
from types import SimpleNamespace

from .dupes import dupes
from .decl import StructDecl, MemberDecl
from .typedef import TypeDef
from .basetypes import base_types
from .member import MemberDef
from .struct import StructDef, StructType
from .jinja import ctemplate, pytemplate
from .c import Include, SysInclude, RelInclude

__all__ = (
    'NameSpace',
)


class NameSpace:
    __slots__ = (
        '_name',
        '_structs',
    )

    _name: Optional[str]
    _structs: Dict[str, StructDef]

    c = ctemplate.get_template('xpdt.c')
    chdr = ctemplate.get_template('apihdr.c')
    capi = ctemplate.get_template('api.c')
    python = pytemplate.get_template('xpdt.pyt')

    def __init__(self,
                 structs: Tuple[StructDef, ...],
                 name: Optional[str] = None):

        d = dupes((s.name for s in structs))
        if d:
            raise ValueError(f'{name}: Duplicated struct name: '
                             ", ".join(d))

        self._name = name

        # TODO: Allow sub-structs
        self._structs = {s.name: s for s in structs}

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        self._name = name

    @property
    def has_name(self) -> bool:
        return self._name is not None

    def __iter__(self) -> Iterator[StructDef]:
        yield from self._structs.values()

    def __len__(self) -> int:
        return len(self._structs)

    @property
    def python_code(self) -> str:
        return self.python.render(namespace=self)

    @property
    def python_type_name(self) -> str:
        name = self._name
        if name is None:
            return 'XpdtBase'
        return f'{name.title()}Type'

    @property
    def python_enum_name(self) -> str:
        name = self._name
        assert name is not None
        return name.title()

    def gen_dynamic_python(self) -> SimpleNamespace:
        names = {s.name for s in self}
        names.add(self.python_type_name)
        if self._name is not None:
            names.add(self.python_enum_name)

        g: Dict[str, Any] = {}
        exec(self.python_code, g, g)
        return SimpleNamespace(**{x: g[x] for x in names})

    def gen_python(self, f: TextIO) -> None:
        f.write(self.python_code)

    def gen_c(self, f: TextIO) -> None:
        headers = (
            SysInclude('stdlib.h'),
            SysInclude('xpdt/xpdt.h'),
        )

        code = self.c.render(
            headers=headers,
            namespace=self,
        )
        f.write(code)

    def gen_c_api_entries(self, f: TextIO,
                          inc_prefix: Optional[str] = None,
                          ) -> None:
        api_hdr: Include
        if not inc_prefix:
            api_hdr = RelInclude(f'{self.name}_api.h')
        else:
            api_hdr = SysInclude(f'{inc_prefix}/{self.name}_api.h')

        headers = (
            api_hdr,
            RelInclude(f'{self.name}_impl.h'),
        )

        code = self.capi.render(
            headers=headers,
            namespace=self,
        )
        f.write(code)

    def gen_c_api_hdr(self, f: TextIO) -> None:
        headers = (
            SysInclude('xpdt/xpdt.h'),
        )
        code = self.chdr.render(
            headers=headers,
            namespace=self,
        )
        f.write(code)

    @classmethod
    def from_decls(cls,
                   decls: Iterable[StructDecl],
                   name: Optional[str] = None) -> 'NameSpace':
        this = name if name is not None else '<anon>'

        typemap = base_types()

        def define_member(decl: MemberDecl) -> MemberDef:
            type_name = decl.type_name
            try:
                member_type = typemap[type_name]
            except KeyError:
                raise KeyError(f'{this}: Type {type_name!r} not known')
            return MemberDef(
                decl.member_name,
                member_type,
            )

        def define_struct(decl: StructDecl) -> StructDef:
            name = decl.struct_name
            struct = StructDef(
                name,
                tuple((define_member(m) for m in decl.members)),
                discriminant=decl.discriminant,
            )
            if name in typemap:
                raise ValueError(f'{this}: Type {name!r} already defined')
            typemap[name] = TypeDef(name, StructType(struct))
            return struct

        defns = tuple((define_struct(decl) for decl in decls))

        return cls(defns, name)



from typing import NamedTuple, List, Optional

__all__ = (
    'MemberDecl',
    'StructDecl',
)


class MemberDecl(NamedTuple):
    member_name: str
    type_name: str


class StructDecl(NamedTuple):
    struct_name: str
    members: List[MemberDecl]
    discriminant: Optional[int] = None

    def with_prefix(self, pfx: str) -> 'StructDecl':
        new_name = f'{pfx}_{self.struct_name}'
        return self._replace(struct_name=new_name)

    def with_discriminant(self, d: int) -> 'StructDecl':
        assert self.discriminant is None
        return self._replace(discriminant=d)



from typing import Tuple, Generator, Iterator, Dict, Optional

from .dupes import dupes

from .type import XpdtType, ConstructAction, ConstructElement
from .member import MemberDef


__all__ = (
    'StructDef',
    'StructType',
)


class StructDef:
    __slots__ = (
        '_name',
        '_discr',
        '_membs',
        '_vmembs',
    )

    _name: str
    _discr: Optional[int]
    _membs: Dict[str, MemberDef]
    _vmembs: Tuple[MemberDef, ...]

    def __init__(self,
                 name: str,
                 members: Tuple[MemberDef, ...],
                 discriminant: Optional[int] = None):
        d = dupes((m.name for m in members))
        if d:
            raise ValueError(f'{name}: Duplicated member name: '
                             ", ".join(d))
        self._name = name
        self._discr = discriminant
        self._membs = {m.name: m for m in members}
        self._vmembs = tuple((m for m in members
                              if m.needs_vbuf))

    @property
    def name(self) -> str:
        return self._name

    @property
    def discriminant(self) -> int:
        assert self._discr is not None
        return self._discr

    @property
    def has_discriminant(self) -> bool:
        return self._discr is not None

    @property
    def ctype(self) -> str:
        return f'struct {self._name}'

    @property
    def member_types(self) -> Generator[XpdtType, None, None]:
        yield from (m.typedef.type for m in self)

    @property
    def vbuf_members(self) -> Tuple[MemberDef, ...]:
        return self._vmembs

    @property
    def non_reserved_vbuf_members(self) -> Generator[MemberDef, None, None]:
        return (m for m in self._vmembs if not m.is_reserved)

    @property
    def non_reserved_members(self) -> Generator[MemberDef, None, None]:
        yield from (m for m in self._membs.values() if not m.is_reserved)

    @property
    def needs_vbuf(self) -> bool:
        return any((t.needs_vbuf for t in self.member_types))

    @property
    def all_members_scalar(self) -> bool:
        return all((t.is_scalar and not t.needs_decode
                    for t in self.member_types))

    @property
    def all_members(self) -> Generator[ConstructElement, None, None]:
        return (elem for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER)

    @property
    def all_non_reserved_members(self,
                                 ) -> Generator[ConstructElement, None, None]:
        return (elem for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER)

    @property
    def all_vbuf_members(self) -> Generator[ConstructElement, None, None]:
        return (elem for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER
                and elem.member.type.needs_vbuf)

    @property
    def scalar_members(self) -> Generator[ConstructElement, None, None]:
        return (elem for elem in
                self.construct_recursive()
                if elem.action == ConstructAction.MEMBER
                and elem.member.is_scalar)

    @property
    def python_var_names(self) -> Generator[str, None, None]:
        return (elem.python_var_name for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER)

    @property
    def c_named_initializers(self) -> Generator[str, None, None]:
        return (elem.c_named_initializer for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER)

    @property
    def c_named_initializers_x1v(self) -> Generator[str, None, None]:
        return (elem.c_named_initializer for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER
                and elem.member.type.needs_vbuf)

    @property
    def python_vbuf_names(self) -> Generator[str, None, None]:
        return (elem.python_var_name for elem in
                self.construct_recursive(include_reserved=False)
                if elem.action == ConstructAction.MEMBER
                and elem.member.type.needs_vbuf)

    def construct_recursive(self,
                            path: Tuple[MemberDef, ...] = (),
                            include_reserved: bool = True,
                            ) -> Generator[ConstructElement, None, None]:
        for m in self:
            if not include_reserved and m.is_reserved:
                continue

            if m.is_scalar:
                yield ConstructElement(ConstructAction.MEMBER, m, path)
            else:
                stype = m.type
                new_path = path + (m,)
                yield ConstructElement(ConstructAction.AGG_PUSH, m, path)
                yield from stype.construct_recursive(
                    new_path,
                    include_reserved=include_reserved,
                )
                yield ConstructElement(ConstructAction.AGG_POP, m, path)

    @property
    def struct_fmt(self) -> str:
        def fmt(m: MemberDef) -> str:
            t = m.typedef.type
            if m.is_reserved:
                return f'{t.size}x'
            else:
                return t.struct_fmt

        return ''.join(map(fmt, self))

    def __getitem__(self, k: str) -> MemberDef:
        return self._membs[k]

    def __iter__(self) -> Iterator[MemberDef]:
        return iter(self._membs.values())

    def __len__(self) -> int:
        return len(self._membs)


class StructType(XpdtType):
    __slots__ = (
        '_struct',
    )

    _struct: StructDef

    def __init__(self, struct: StructDef):
        self._struct = struct

    @property
    def _members(self) -> Generator[MemberDef, None, None]:
        yield from self._struct

    @property
    def _member_types(self) -> Generator[XpdtType, None, None]:
        yield from (m.typedef.type for m in self._struct)

    @property
    def struct_fmt(self) -> str:
        return self._struct.struct_fmt

    @property
    def struct(self) -> StructDef:
        return self._struct

    @property
    def ctype(self) -> str:
        return self._struct.ctype

    @property
    def needs_vbuf(self) -> bool:
        return any((t.needs_vbuf for t in self._member_types))

    @property
    def needs_decode(self) -> bool:
        raise NotImplementedError

    @property
    def pytype(self) -> str:
        return self._struct.name

    @property
    def is_scalar(self) -> bool:
        return False

    @property
    def struct_name(self) -> str:
        return self._struct.name

    def construct_recursive(self,
                            path: Tuple[MemberDef, ...] = (),
                            include_reserved: bool = True,
                            ) -> Generator[ConstructElement, None, None]:
        yield from self._struct.construct_recursive(
            path,
            include_reserved=include_reserved,
        )

    @property
    def scalar_members(self) -> Generator[ConstructElement, None, None]:
        yield from self._struct.scalar_members



from typing import ClassVar, TextIO, Optional
from argparse import ArgumentParser
from pathlib import Path
from sys import stdout
import logging

from xpdt import NameSpace, load

_fmt = logging.Formatter('%(message)s')
_stdio_handler = logging.StreamHandler(stream=stdout)
_stdio_handler.setFormatter(_fmt)
_log = logging.getLogger('xpdt')
_log.addHandler(_stdio_handler)


class BackendDef:
    suffix: ClassVar[str]

    __slots__ = (
        '_p',
        '_ns',
        '_inc_prefix',
    )

    _p: Path
    _ns: NameSpace
    _inc_prefix: Optional[str]

    def _name(self, base: Path, name: str) -> Path:
        return base / (name + self.suffix)

    def __init__(self, ns: NameSpace, base: Path,
                 inc_prefix: Optional[str] = None) -> None:
        self._ns = ns
        name = ns.name
        assert name is not None
        self._p = self._name(base, name)
        self._inc_prefix = inc_prefix

    def _gen(self, f: TextIO) -> None:
        raise NotImplementedError

    def gen(self) -> None:
        _log.info('Writing code: %s', self._p)
        with self._p.open('w') as f:
            return self._gen(f)


class OutputC(BackendDef):
    suffix = '_impl.h'
    __slots__ = ()

    def _gen(self, f: TextIO) -> None:
        self._ns.gen_c(f)


class OutputCAPI(BackendDef):
    suffix = '.c'
    __slots__ = ()

    def _gen(self, f: TextIO) -> None:
        self._ns.gen_c_api_entries(f, self._inc_prefix)


class OutputCHdr(BackendDef):
    suffix = '_api.h'
    __slots__ = ()

    def _gen(self, f: TextIO) -> None:
        self._ns.gen_c_api_hdr(f)


class OutputPy(BackendDef):
    suffix = '.py'
    __slots__ = ()

    def _gen(self, f: TextIO) -> None:
        self._ns.gen_python(f)


def main() -> None:
    opts = ArgumentParser(description='xpdt: eXPeditious Data Transfer')
    opts.add_argument('--verbose', '-v',
                      action='count',
                      default=0,
                      help='Be more talkative')
    opts.add_argument('--registry', '-r',
                      default=None,
                      type=Path,
                      help='Path to registry.toml')
    opts.add_argument('--module-name', '-n',
                      default=None,
                      type=str,
                      help='Name of module (output file)')
    opts.add_argument('--language', '-l',
                      default='c',
                      type=str,
                      help='Output language')
    opts.add_argument('--out', '-o',
                      default=Path(),
                      type=Path,
                      help='Output dir')
    opts.add_argument('-I',
                      dest='inc_prefix',
                      help='Include path (for C headers)')
    opts.add_argument('file',
                      type=Path,
                      nargs='+',
                      help='xpdt schema files')

    args = opts.parse_args()
    if args.verbose:
        _log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)

    ns = load(
        args.file,
        registry=args.registry,
        module_name=args.module_name,
    )

    if ns.name is None:
        ns.name = 'xpdt'

    backends = {
        'c': OutputC,
        'capi': OutputCAPI,
        'chdr': OutputCHdr,
        'py': OutputPy,
        'python': OutputPy,
    }

    try:
        cls = backends[args.language]
    except KeyError:
        print(f'Unknown language: "{args.language}"')
        raise SystemExit(1)

    args.out.mkdir(exist_ok=True)
    backend = cls(ns, args.out, args.inc_prefix)
    backend.gen()


if __name__ == '__main__':
    main()



from pathlib import PurePosixPath

__all__ = (
    'Include',
    'SysInclude',
    'RelInclude',
)


class Include(PurePosixPath):
    @property
    def path_spec(self) -> str:
        raise NotImplementedError

    @property
    def include(self) -> str:
        return f'#include {self.path_spec}'


class SysInclude(Include):
    @property
    def path_spec(self) -> str:
        return f'<{self}>'


class RelInclude(Include):
    @property
    def path_spec(self) -> str:
        return f'"{self}"'
